package org.nakedobjects.nof.boot.system;

import org.nakedobjects.nof.core.conf.Configuration;
import org.nakedobjects.nof.core.context.NakedObjectsContext;

import java.io.IOException;
import java.io.PrintWriter;
import java.net.URLDecoder;
import java.util.StringTokenizer;


public class HttpServerMonitor extends AbstractServerMonitor {
    private static final int DEFAULT_PORT = 8081;
    private static final String PORT = Configuration.ROOT + "monitor.http.port";
    private final Monitor monitor = new Monitor();

    protected int getPort() {
        return NakedObjectsContext.getConfiguration().getInteger(PORT, DEFAULT_PORT);
    }

    protected void handleRequest(PrintWriter writer, String request) throws IOException {
        StringTokenizer st = new StringTokenizer(request);
        if (st.countTokens() != 3) {
            httpErrorResponse(writer, 444, "Unparsable input " + request);
            return;
        }

        String type = st.nextToken();
        if (!type.equals("GET")) {
            httpErrorResponse(writer, 400, "Invalid method " + type);
            return;
        }
        
        writer.println("HTTP/1.0 200");
        writer.println("Content-Type: text/html");
        writer.println("");

        String query = st.nextToken();
        query = URLDecoder.decode(query, "UTF-8");
        
        if (query.startsWith("/monitor")) {
            query = query.substring("/monitor".length());
            
            if (query.startsWith("?")) {
                query = query.substring(1);
            }
        
            monitor.writePage(query, writer);
        } else {
            httpErrorResponse(writer, 404, "Failed to find " + query);
        }
    }

    private void httpErrorResponse(PrintWriter writer, int errorNo, String response) {
        writer.println("HTTP/1.0 " + errorNo + " " + response);
        writer.println("Content-Type: text/html");
        writer.println("");

        writer.println("<HTML><HEAD><TITLE>Error " + errorNo + " - " + response + "</TITLE></HEAD>");
        writer.println("<BODY><h1>" + errorNo + " - " + response + "</h1>");
        writer.println("</BODY></HTML>");

        writer.flush();
    }
    
    public void setTarget(NakedObjectsSystem system) {
        monitor.setTarget(system);
    }

}

// Copyright (c) Naked Objects Group Ltd.
