package org.nakedobjects.nof.boot.system;

import java.io.File;
import java.io.FilenameFilter;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Locale;
import java.util.Properties;
import java.util.TimeZone;
import java.util.Vector;

import org.apache.log4j.Appender;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Layout;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.PropertyConfigurator;
import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedObjectLoader;
import org.nakedobjects.noa.adapter.Oid;
import org.nakedobjects.noa.adapter.ResolveState;
import org.nakedobjects.noa.client.NakedObjectClient;
import org.nakedobjects.noa.persist.NakedObjectPersistor;
import org.nakedobjects.noa.reflect.NakedObjectReflector;
import org.nakedobjects.noa.security.AuthenticationRequest;
import org.nakedobjects.noa.security.ConnectionManager;
import org.nakedobjects.noa.security.Session;
import org.nakedobjects.noa.security.SessionManager;
import org.nakedobjects.nof.core.client.ClientInstaller;
import org.nakedobjects.nof.core.conf.Configuration;
import org.nakedobjects.nof.core.conf.DefaultConfigurationLoader;
import org.nakedobjects.nof.core.context.ContextDebug;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.image.SplashWindow;
import org.nakedobjects.nof.core.image.TemplateImageLoader;
import org.nakedobjects.nof.core.security.AuthenticatorInstaller;
import org.nakedobjects.nof.core.service.ServiceUtil;
import org.nakedobjects.nof.core.system.ClientConnectionInstaller;
import org.nakedobjects.nof.core.system.ConfigurationLoader;
import org.nakedobjects.nof.core.system.FixturesInstaller;
import org.nakedobjects.nof.core.system.ObjectLoaderInstaller;
import org.nakedobjects.nof.core.system.ObjectPersistorInstaller;
import org.nakedobjects.nof.core.system.ReflectorEnhancementInstaller;
import org.nakedobjects.nof.core.system.ReflectorInstaller;
import org.nakedobjects.nof.core.system.ServerListener;
import org.nakedobjects.nof.core.system.ServerListenerInstaller;
import org.nakedobjects.nof.core.system.ServicesInstaller;
import org.nakedobjects.nof.core.system.TemplateImageLoaderInstaller;
import org.nakedobjects.nof.core.util.AboutNakedObjects;
import org.nakedobjects.nof.core.util.Assert;
import org.nakedobjects.nof.core.util.DebugInfo;
import org.nakedobjects.nof.core.util.DebugInfo2;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.core.util.InfoDebugFrame;
import org.nakedobjects.nof.core.util.NakedObjectConfiguration;


/**
 * TODO move this to org.nakedobjects.nof.core.system in core project once reflector dependency has been
 * removed
 */
public final class NakedObjectsSystem implements ConnectionManager, DebugInfo2 {
    private final static class InitialisationSession implements Session {
        private static final long serialVersionUID = 1L;

        public byte[] asEncodedString() {
            return null;
        }

        public String[] getRoles() {
            return new String[0];
        }

        public String getUserName() {
            return "initialisation";
        }

        public String getValidationCode() {
            return "";
        }

        public void restoreFromEncodedString(byte[] data) {}
    }

    private static final Logger LOG = Logger.getLogger(NakedObjectsSystem.class);

    private final static String LOGGING_CONFIG_FILE = "logging.properties";

    private static void print(String string) {
        System.out.println(string);
    }

    private static void printBlock(String title) {
        print("");
        print("------------------------------------------");
        print(title);
        print("------------------------------------------");
    }

    public static void printDiagnostics() {
        print("------- Naked Objects diagnostics report -------");
        printVersion();

        printBlock("System properties");
        Properties properties = System.getProperties();
        Enumeration names = properties.propertyNames();
        while (names.hasMoreElements()) {
            String name = (String) names.nextElement();
            print(name + "= " + properties.getProperty(name));
        }

        File file = new File("../lib");
        if (file.isDirectory()) {
            String[] files = file.list(new FilenameFilter() {
                public boolean accept(File dir, String name) {
                    return name.endsWith(".jar");
                }
            });
            printBlock("Libs");
            for (int i = 0; i < files.length; i++) {
                print(files[i]);
            }
        }

        file = new File("config");
        if (file.isDirectory()) {
            String[] files = file.list(new FilenameFilter() {
                public boolean accept(File dir, String name) {
                    return true;
                }
            });
            printBlock("Config files");
            for (int i = 0; i < files.length; i++) {
                print(files[i]);
            }
        }

        printBlock("Locale information");
        print("Default locale: " + Locale.getDefault());
        print("Default timezone: " + TimeZone.getDefault());
    }

    public static void printVersion() {
        String date = AboutNakedObjects.getFrameworkCompileDate();
        String compileDate = date == null ? "" : ", compiled on " + date;
        print(AboutNakedObjects.getFrameworkName() + ", " + AboutNakedObjects.getFrameworkVersion() + compileDate);
    }

    private NakedObjectConfiguration configuration;
    private FixturesInstaller fixtureIinstaller;
    private ObjectLoaderInstaller objectLoaderInstaller;
    private ObjectPersistorInstaller persistorInstaller;
    private AuthenticatorInstaller authenticatorInstaller;
    private Vector serverListenerInstallers = new Vector();;
    private ServicesInstaller servicesInstaller;
    private boolean shareObjectsAcrossContexts;
    private boolean shareObjectsAcrossContextsIsSet = false;
    private Object[] services;
    private boolean hideSplash = false;
    private NakedObjectReflector reflector;
    private NakedObjectLoader sharedObjectLoader;
    private NakedObjectPersistor sharedObjectPersistor;
    private SplashWindow splash;
    private TemplateImageLoader templateImageLoader;
    private Vector listeners = new Vector();
    private SessionManager sessionManager;
    private InfoDebugFrame debugFrame;
    private InstallerLookup installerLookup;
    private ReflectorInstaller reflectorInstaller;
    private NakedObjectsContext context;
    private ConnectionManager connectionManager;
    private boolean isServer;

    public NakedObjectsSystem() {
        this(Configuration.DIRECTORY);
    }

    public NakedObjectsSystem(final String configDirectory) {
        this(configDirectory, null);
    }

    public NakedObjectsSystem(final String configDirectory, final Level level) {
        /*
         * this sets up logging temporarily until the required logging details are loaded by the
         * PropertyConfigurator in installConfiguration().
         */
        File f = new File(configDirectory, LOGGING_CONFIG_FILE);
        if (f.exists()) {
            PropertyConfigurator.configure(f.getAbsolutePath());
        } else {
            Layout layout = new PatternLayout("%-5r [%-25.25c{1} %-10.10t %-5.5p]  %m%n");
            Appender appender = new ConsoleAppender(layout);
            //appender.addFilter(new LevelMatchFilter() {});
            BasicConfigurator.configure(appender);
            Logger.getRootLogger().setLevel(Level.WARN);
            Logger.getLogger("ui").setLevel(Level.OFF);
        }

        if (level != null) {
            Logger.getRootLogger().setLevel(level);
        }
        installerLookup = new InstallerLookup();
    }

    private void adaptServices() {
        NakedObjectPersistor persistor = NakedObjectsContext.getObjectPersistor();
        NakedObjectReflector specLoader = NakedObjectsContext.getReflector();
        NakedObjectLoader loader = NakedObjectsContext.getObjectLoader();
        for (int i = 0; i < services.length; i++) {
            specLoader.installServiceSpecification(services[i].getClass());
            Oid oid = persistor.getOidForService(ServiceUtil.id(services[i]));
            if (oid == null) {
                throw new NakedObjectRuntimeException();
/*                NakedObject adapter = loader.createAdapterForTransient(services[i], false);
                loader.madePersistent(adapter);
                oid = adapter.getOid();
                persistor.registerService(ServiceUtil.id(services[i]), oid);
   */
            } else {
                NakedObject adapter = loader.recreateAdapter(oid, services[i]);
                if (adapter.getResolveState().isResolvable(ResolveState.RESOLVING)) {
                    adapter.changeState(ResolveState.RESOLVING);
                    adapter.changeState(ResolveState.RESOLVED);
                }
            }
        }
    }

    public void addListener(final ServerListenerInstaller installer) {
        serverListenerInstallers.add(installer);
    }

    public void startListening() {
        final Iterator iter = serverListenerInstallers.iterator();
        while (iter.hasNext()) {
            final ServerListenerInstaller serverListenerInstaller = (ServerListenerInstaller) iter.next();
            final ServerListener listener = serverListenerInstaller.createListener(configuration);
            listeners.add(listener);
            listener.setConnectionManager(this);
            listener.setSessionManager(sessionManager);
            Thread thread = new Thread() {
                public void run() {
                    listener.listen();
                }
            };
            thread.start();
        }
    }

    public void connect(Session session) {
        initContext(session, shareObjectsAcrossContexts);
    }

    private void debug(final DebugString debug, final Object object) {
        if (object instanceof DebugInfo) {
            DebugInfo d = (DebugInfo) object;
            debug.appendTitle(d.debugTitle());
            d.debugData(debug);
        } else {
            debug.appendln(object.toString());
            debug.appendln("... no further debug information");
        }
    }

    public void debug(final String selector, final DebugString debug) {
        if (selector.equals("Overview")) {
            debugOverview(debug);
        } else if (selector.equals("Configuration")) {
            debug(debug, configuration);
        } else if (selector.equals("Authenticator")) {
            debug(debug, sessionManager);
        } else if (selector.equals("Reflector")) {
            debug(debug, reflector);
        } else if (selector.equals("Contexts")) {
            debugListContexts(debug);
        } else if (selector.equals("Listeners")) {
            debugListeners(debug);
        } else if (selector.equals("Events")) {

            debugListeners(debug);
        } else {
            debugDisplayContext(selector, debug);
        }
    }

    private void debugDisplayContext(final String selector, final DebugString debug) {
        ContextDebug d = NakedObjectsContext.getDebugContext(selector);
        if (d != null) {
            d.debugAll(debug);
        } else {
            debug.appendln("No context: " + selector);
        }
    }

    private void debugListContexts(final DebugString debug) {
        String[] contextIds = NakedObjectsContext.getAllExecutionContextIds();
        for (int i = 0; i < contextIds.length; i++) {
            debug.appendln(contextIds[i]);
            debug.appendln("-----");
            ContextDebug d = NakedObjectsContext.getDebugContext(contextIds[i]);
            d.debug(debug);
            debug.appendln();
        }
    }

    private void debugListeners(DebugString debug) {
        Enumeration e = listeners.elements();
        while (e.hasMoreElements()) {
            debug(debug, e.nextElement());
        }
    }

    public String[] debugOptions() {
        String[] general = new String[] { "Overview", "Authenticator", "Configuration", "Reflector", "Listeners", "Events",
                "Contexts" };
        String[] contextIds = NakedObjectsContext.getAllExecutionContextIds();
        String[] combined = new String[general.length + contextIds.length];
        System.arraycopy(general, 0, combined, 0, general.length);
        System.arraycopy(contextIds, 0, combined, general.length, contextIds.length);
        return combined;
    }

    private void debugOverview(final DebugString debug) {
        debug.appendln(AboutNakedObjects.getFrameworkName());
        debug.appendln(AboutNakedObjects.getFrameworkVersion());
        if (AboutNakedObjects.getApplicationName() != null) {
            debug.appendln("application: " + AboutNakedObjects.getApplicationName());
        }
        if (AboutNakedObjects.getApplicationVersion() != null) {
            debug.appendln("version" + AboutNakedObjects.getApplicationVersion());
        }

        String user = System.getProperty("user.name");
        String system = System.getProperty("os.name") + " (" + System.getProperty("os.arch") + ") "
                + System.getProperty("os.version");
        String java = System.getProperty("java.vm.name") + " " + System.getProperty("java.vm.version");
        debug.appendln("user: " + user);
        debug.appendln("os: " + system);
        debug.appendln("java: " + java);
        debug.appendln("working directory: " + new File(".").getAbsolutePath());

        debug.appendTitle("System Installer");
        debug.appendln("Fixture Installer", fixtureIinstaller.getClass().getName());
        debug.appendln("Object Loader Installer", objectLoaderInstaller.getClass().getName());

        debug.appendTitle("System Components");
        debug.appendln("Authentication manager", sessionManager.getClass().getName());
        debug.appendln("Configuration", configuration.getClass().getName());

        DebugInfo[] inf = NakedObjectsContext.debug();
        for (int i = 0; i < inf.length; i++) {
            if (inf[i] != null) {
                inf[i].debugData(debug);
            }
        }
    }

    public void disableSplash(boolean hideSplash) {
        this.hideSplash = hideSplash;
    }

    public void disconnect(Session session) {

        /*
         * 1. confirm that session exist (user is connected)
         * 
         * 2. remove session from pool singleton, and get singleton reference for this thread to be disposed.
         */
        resetContext();
    }

    public SessionManager getAuthenticationManager() {
        return sessionManager;
    }

    public InstallerLookup getInstallerFactory() {
        return installerLookup;
    }

    public void init() {
        LOG.info("initialising naked objects system");
        LOG.info("working directory: " + new File(".").getAbsolutePath());
        LOG.info("root path: " + NakedObjectsContext.getConfiguration().rootPath());
        try {
            initLocale();
            initTimeZone();
            initImageLoader();
            showSplash();
            sessionManager = authenticatorInstaller.createAuthenticationManager();
            startListening();
            if (!shareObjectsAcrossContextsIsSet) {
                shareObjectsAcrossContexts = configuration.getBoolean(Configuration.ROOT + "share-objects-across-context", false);
                shareObjectsAcrossContextsIsSet = true;
            }

            String[] peers = configuration.getList(Configuration.ROOT + "reflector.peers");
            for (int i = 0; peers != null && i < peers.length; i++) {
                LOG.info("adding reflector peer " + peers[i]);
                ReflectorEnhancementInstaller enhancement = (ReflectorEnhancementInstaller) installerLookup.getInstaller(
                        peers[i], ReflectorEnhancementInstaller.class);
                reflectorInstaller.addEnhancement(enhancement);
            }

            reflector = reflectorInstaller.createReflector(configuration);
            reflector.init();

            if (objectLoaderInstaller == null) {
                objectLoaderInstaller = persistorInstaller.getPreferredObjectLoaderInstaller();
            }
            if (objectLoaderInstaller == null) {
                objectLoaderInstaller = installerLookup.objectLoaderInstaller("java");
            }

            persistorInstaller.setUpReflector(reflector);

            if (shareObjectsAcrossContexts) {
                sharedObjectLoader = objectLoaderInstaller.createObjectLoader(configuration);
                sharedObjectPersistor = persistorInstaller.createObjectPersistor(configuration);
            }

            // TODO install the strategies for remoting
            // installerLookup.clientConnectionInstaller(null).installStrategies(configuration);

            services = servicesInstaller.getServices(isServer);
            initContext(new InitialisationSession(), shareObjectsAcrossContexts);

            if (shareObjectsAcrossContexts) {
                sharedObjectLoader.setServices(services);
                sharedObjectLoader.init();
                sharedObjectPersistor.setServices(services);
                sharedObjectPersistor.init();
                adaptServices();
            }

            if (fixtureIinstaller != null) {
                fixtureIinstaller.installFixtures(configuration);
            }

            // remove all unneeded details in the initialisation context
            NakedObjectsContext.getMessageBroker().getWarnings();
            NakedObjectsContext.getMessageBroker().getMessages();
            NakedObjectsContext.getUpdateNotifer().allChangedObjects();
            NakedObjectsContext.getUpdateNotifer().allDisposedObjects();
            resetContext();

            // TODO shutdown the startup context (eg on main thread)
            // NakedObjectsContext.shutdown();
            sessionManager.init();

        } catch (RuntimeException e) {
            LOG.error("failed to intialise", e);
            removeSplash(0);
            throw e;
        }
    }

    private void initContext(final Session session, final boolean shareObjectsAcrossContexts) {
        if (NakedObjectsContext.getSession() != null) {
            throw new NakedObjectRuntimeException("Session already exists for naked objects context: "
                    + NakedObjectsContext.getSession() + " on " + Thread.currentThread());
        }
        context.setSession(session);
        if (NakedObjectsContext.isReady()) {
            if (!shareObjectsAcrossContexts) {
                adaptServices();
            }
        } else {
            context.setTemplateImageLoader(templateImageLoader);
            context.setReflector(reflector);
            if (shareObjectsAcrossContexts) {
                context.setObjectLoader(sharedObjectLoader);
                context.setObjectPersistor(sharedObjectPersistor);
            } else {
                final NakedObjectLoader objectLoader = objectLoaderInstaller.createObjectLoader(configuration);
                final NakedObjectPersistor objectPersistor = persistorInstaller.createObjectPersistor(configuration);
                context.setObjectLoader(objectLoader);
                context.setObjectPersistor(objectPersistor);
                objectLoader.setServices(services);
                objectLoader.init();
                objectPersistor.setServices(services);
                objectPersistor.init();
                adaptServices();
            }
        }

    }

    private void initImageLoader() {
        Assert.assertNotNull(templateImageLoader);
        templateImageLoader.init();
    }

    private void initLocale() {
        String localeSpec = configuration.getString(Configuration.ROOT + "locale");
        if (localeSpec != null) {
            int pos = localeSpec.indexOf('_');
            Locale locale;
            if (pos == -1) {
                locale = new Locale(localeSpec, "");
            } else {
                String language = localeSpec.substring(0, pos);
                String country = localeSpec.substring(pos + 1);
                locale = new Locale(language, country);
            }
            Locale.setDefault(locale);
            LOG.info("locale set to " + locale);
        }
        LOG.debug("locale is " + Locale.getDefault());
    }

    private void initTimeZone() {
        String timeZoneSpec = configuration.getString(Configuration.ROOT + "timezone");
        if (timeZoneSpec != null) {
            TimeZone timeZone;
            timeZone = TimeZone.getTimeZone(timeZoneSpec);
            TimeZone.setDefault(timeZone);
            LOG.info("time zone set to " + timeZone);
        }
        LOG.debug("time zone is " + TimeZone.getDefault());
    }

    public void loadConfiguration(final ConfigurationLoader installer) {
        ConfigurationLoader installConfiguration = installer == null ? new DefaultConfigurationLoader() : installer;
        configuration = installConfiguration.load();
        NakedObjectsContext.setConfiguration(configuration);
    }

    public void removeSplash() {
        removeSplash(6);
    }

    private void removeSplash(int delay) {
        if (splash != null) {
            if (delay == 0) {
                splash.removeImmediately();
            } else {
                splash.toFront();
                splash.removeAfterDelay(delay);
            }
        }
    }

    private void resetContext() {
        context.clearSession();

        if (!shareObjectsAcrossContexts) {
            NakedObjectsContext.getObjectLoader().reset();
            NakedObjectsContext.getObjectPersistor().reset();
        }
    }

    public void runClient(
            final ClientInstaller installer,
            final String displayTitle,
            final boolean authenticateUser,
            final boolean showExplorationOptions,
            final AuthenticationRequest authenticationRequest) {

        if (connectionManager == null) {
            connectionManager = this;
        }

        try {
            NakedObjectClient client = installer.createClient(configuration);
            client.setTitle(displayTitle);
            client.setExploration(showExplorationOptions);
            client.setAuthenticate(authenticateUser);
            client.setAuthenticationRequestViaArgs(authenticationRequest);
            client.setConnectionManager(connectionManager);
            client.setSessionManager(sessionManager);
            client.run();
        } catch (Exception e) {
            LOG.error("client start up problem", e);
            removeSplash(0);
        }
    }

    public void setAuthenticatorInstaller(final AuthenticatorInstaller installAuthenticator) {
        authenticatorInstaller = installAuthenticator;
    }

    /**
     * Set the type of connection to used to access the server.
     */
    public void setClientConnection(final ClientConnectionInstaller installer) {
        setObjectPersistorInstaller(installer);
        reflectorInstaller.addEnhancement(installer);
        setAuthenticatorInstaller(installer);
    }

    public void setContext(final NakedObjectsContext context) {
        this.context = context;
    }

    public void setFixtureInstaller(final FixturesInstaller installer) {
        this.fixtureIinstaller = installer;
    }

    public void setObjectLoaderInstaller(final ObjectLoaderInstaller factory) {
        this.objectLoaderInstaller = factory;
    }

    public void setObjectPersistorInstaller(final ObjectPersistorInstaller factory) {
        this.persistorInstaller = factory;
    }

    public void setReflectorInstaller(ReflectorInstaller reflectorInstaller) {
        this.reflectorInstaller = reflectorInstaller;
    }

    public void setServer(boolean isServer) {
        this.isServer = isServer;
    }

    public void setServicesInstaller(final ServicesInstaller installer) {
        this.servicesInstaller = installer;
    }

    public void setShareObjectsAcrossContexts(final boolean shareObjectsAcrossContexts) {
        this.shareObjectsAcrossContexts = shareObjectsAcrossContexts;
        shareObjectsAcrossContextsIsSet = true;
    }

    public void setTemplateImageLoader(final TemplateImageLoaderInstaller loader) {
        this.templateImageLoader = loader.createLoader();
    }

    private void showSplash() {
        boolean noSplash = configuration.getBoolean(Configuration.ROOT + "nosplash", false) || hideSplash;
        if (!noSplash) {
            splash = new SplashWindow(templateImageLoader);
        }
    }

    public void shutdown() {
        LOG.info("shutting down system");
        if (debugFrame != null) {
            debugFrame.dispose();
            debugFrame = null;
        }
        Enumeration e = listeners.elements();
        while (e.hasMoreElements()) {
            ServerListener listener = (ServerListener) e.nextElement();
            listener.stop();
        }
        NakedObjectsContext.shutdown();
        reflector.shutdown();
    }
}
// Copyright (c) Naked Objects Group Ltd.
